import { intrinsic, rootSchema } from "@ark/schema";
import { arkModule } from "../utils.js";
const jsonStringDescription = "a JSON string";
const isParsableJson = (s) => {
    try {
        JSON.parse(s);
        return true;
    }
    catch {
        return false;
    }
};
const root = rootSchema({
    domain: "string",
    predicate: {
        meta: jsonStringDescription,
        predicate: isParsableJson
    }
});
export const json = arkModule({
    root,
    parse: rootSchema({
        in: "string",
        morphs: (s, ctx) => {
            if (s.length === 0) {
                return ctx.error({
                    code: "predicate",
                    expected: jsonStringDescription,
                    actual: "empty"
                });
            }
            try {
                return JSON.parse(s);
            }
            catch (e) {
                return ctx.error({
                    code: "predicate",
                    expected: jsonStringDescription,
                    problem: `must be ${jsonStringDescription} (${e})`
                });
            }
        },
        declaredOut: intrinsic.json
    })
});
